#ifndef COMMANDS_MONITOR_H
#define COMMANDS_MONITOR_H

#include <iostream>
#include <chrono>
#include <iomanip>

#include <string>
#include <kr2_program_api/api_v1/bundles/custom_device.h>
#include <kr2/utils/stats.h>
#include <kr2/utils/configuration_file.h>
#include "kr2/kord/robot_control_action.h"
#include <boost/property_tree/ini_parser.hpp>
#include "kr2_rc_api/internals/console.h"

namespace kr2 {
namespace utils {
    inline timespec chronoToTimespec(const std::chrono::system_clock::time_point& a_time_point){
        auto seconds = std::chrono::duration_cast<std::chrono::seconds>(a_time_point.time_since_epoch());
        auto nanoseconds = std::chrono::duration_cast<std::chrono::nanoseconds>(a_time_point.time_since_epoch()) -
                       std::chrono::duration_cast<std::chrono::nanoseconds>(seconds);
 
        return { seconds.count(), nanoseconds.count() };
    }
}
}

class CommandsMonitor {
	public:
        CommandsMonitor(boost::shared_ptr<kr2_program_api::ProgramInterface> a_api, ConfifurationFile::PARSED_DATA a_data)
        :accum_ltick{Accum<int64_t>(a_data.STATS_INTERVAL / KORD_PERIOD_)}, accum_stick{Accum<int64_t>(a_data.STATS_INTERVAL / KORD_PERIOD_)},
        accum_wake_up_stats{Accum<int64_t>(a_data.STATS_INTERVAL / KORD_PERIOD_)}{
            program_api_ = a_api;
            stats_parameters_ = a_data;

            reset();
        }

        // for joints
        void calculateContinousMoveStats(std::chrono::_V2::system_clock::time_point a_recent_capt_timestamp, timespec a_recent_capt_content_timestamp,
            kr2::kord::MoveBase<double, 7UL>& a_move){
            double sum_angular_diff = 0;

            for (std::size_t i = 0; i < a_move.getTargets().size(); ++i) {
                sum_angular_diff += abs(a_move.getTargets()[i] - last_joints_target[i]);
            }

            bool n = calculateCaptureStats(true, sum_angular_diff, a_recent_capt_timestamp, a_recent_capt_content_timestamp, a_move.getTrackingType(), a_move.getTrackingTypeValue(),
                a_move.getBlendType(), a_move.getBlendTypeValue());
            if (n){
                last_joints_target = a_move.getTargets();
                last_diff = sum_angular_diff;
            }
                
        }

        // for linear
        void calculateContinousMoveStats(std::chrono::_V2::system_clock::time_point a_recent_capt_timestamp, timespec a_recent_capt_content_timestamp,
            kr2::kord::MoveBase<double, 6UL>& a_move){
            double sum_linear_diff = 0;

            for (std::size_t i = 0; i < a_move.getTargets().size(); ++i) {
                sum_linear_diff += abs(a_move.getTargets()[i] - last_linear_target[i]);
            }

            bool n = calculateCaptureStats(false, sum_linear_diff, a_recent_capt_timestamp, a_recent_capt_content_timestamp, a_move.getTrackingType(), a_move.getTrackingTypeValue(),
                a_move.getBlendType(), a_move.getBlendTypeValue());
            if (n){
                last_linear_target = a_move.getTargets();
                last_diff = sum_linear_diff;
            }
                
        }

        // SyncNow sleep jitter monitor, input comes in nanos
        void processNewSystemJitter(const int64_t& a_system_jitter){
            // if (last_received_seq_number_ == -1) // start of recording system jitters after at least one command received
            //     return;
            accum_wake_up_stats.measurement(a_system_jitter / NANOSECS_IN_MICRO_);
            accum_wake_up_stats.compute();
            if (accum_wake_up_stats.get_avg() > stats_parameters_.MAX_RECENT_AVG_SYSTEM_JITTER){
                doCBunError(system_alert_prefix + ": " + "Average Local System Jitter Exceeded: " + doubleToStringWithPrecision(accum_wake_up_stats.get_avg()) + " us, "
                 + "LIMIT: " + std::to_string(stats_parameters_.MAX_RECENT_AVG_SYSTEM_JITTER) + " us");
                 sys_problem = true;
            } 
            if (accum_wake_up_stats.get_max() > stats_parameters_.MAX_RECENT_MAX_SYSTEM_JITTER){
                doCBunError(system_alert_prefix + ": " + "Maximum Local System Jitter Exceeded: " + doubleToStringWithPrecision(accum_wake_up_stats.get_max()) + " us, "
                 + "LIMIT: " + std::to_string(stats_parameters_.MAX_RECENT_MAX_SYSTEM_JITTER) + " us");
                sys_problem = true;
            }

            if (accum_wake_up_stats.get_max() > global_system_jitter_max){ // stored in micros
                global_system_jitter_max = accum_wake_up_stats.get_max();
            }
        }

        void processNewSequenceNumber(const uint16_t& a_rcv_sequence_number){

            checkOfTheReceivedSequenceNumberGlobal(a_rcv_sequence_number);
            checkOfTheReceivedSequenceNumberLocal(a_rcv_sequence_number);

            last_received_seq_number_ = a_rcv_sequence_number;
        }

        void reset(){
            // if (last_received_seq_number_ != -1){
            //     dialog_showed_ = false;
            // }
            //dialog_showed_ = false;
            local_packets_lost_seq = 0;
            local_packets_lost_timestmp = 0;
            global_ltick_max = 0;
            global_packets_lost_seq = 0;
            global_packets_lost_timestmp = 0;
            global_system_jitter_max = 0;
            global_stick_max = 0;
            last_received_seq_number_ = -1;
            clear(stats_buffer_seq);
            clear(stats_buffer_times);
            accum_ltick.reset();
            accum_stick.reset();
            accum_wake_up_stats.reset();
        }

        void setSystemEventState(int a_state){
            if ((dialog_showed_ == true) && is_in_system_event_ && ((a_state == STANDBY_BIT))){ // switch from suspended -> normal
                dialog_showed_ = false;
                
                reset();
            }
            is_in_system_event_  = a_state & SUSPENDED_BIT;
        }

        bool cleanAlarmState(){
            dialog_showed_ = false;
            reset();
            return true;
        }

        bool getAlarmState(){
            return dialog_showed_;
        }

        bool setAlarmState(bool a_state, bool show_unsuspend_error = true){
            
            if (show_unsuspend_error){
                doCBunError(robot_alert_prefix + ": " + "Robot SUSPENDED, all commands are rejected.");
            }
            else{
                dialog_showed_ = a_state;
            }
            return true;
        }

    private:

    std::string system_alert_prefix = "KORD SYSTEM ALERT";
    std::string robot_alert_prefix = "KORD ROBOT ALERT";
    std::string network_alert_prefix = "KORD NETWORK ALERT";

    bool is_in_system_event_ = false;

    boost::shared_ptr<kr2_program_api::ProgramInterface> program_api_;

    int last_received_seq_number_ = -1;
    bool dialog_showed_ = false; // represents alarm state
    bool notified_about_velocity = false;
    //Setup
    //
    const unsigned int NANOSECS_IN_MILLI_ = 1000000;
    const unsigned int MICROSECS_IN_MILLI_ = 1000;
    const unsigned int MILLISECS_IN_SECOND_ = MICROSECS_IN_MILLI_;
    const unsigned int NANOSECS_IN_MICRO_ = MICROSECS_IN_MILLI_;

    const unsigned int KORD_PERIOD_ = 4; //milliseconds

    static constexpr int SUSPENDED_BIT = 32;
    static constexpr int STANDBY_BIT = 1;

    // parameters
    ConfifurationFile::PARSED_DATA stats_parameters_;
    
    std::queue<int> stats_buffer_seq;
    std::queue<int> stats_buffer_times;

    // for captured statistics
    std::chrono::high_resolution_clock::time_point last_timestamp_{};

    // to check 0 velocity
    std::array<double, 7UL> last_joints_target;
    double last_diff;
    std::array<double, 6UL> last_linear_target;

    void doCBunError(std::string a_description){
        if (!dialog_showed_){
            auto dialog = kr2_program_api::OutputDialog::factory("CBun KORD", a_description);
            dialog->setIcon(kr2_program_api::BaseDialog::ERROR);
            //with image
            dialog->show(false);
            dialog_showed_ = true;
            if (stats_parameters_.HALT_TRIGGER) // if HALT_TRIGGER we anyway stop
                stopRobot();
        }
    }

    // checks sequence numbers continuity
    void checkOfTheReceivedSequenceNumberGlobal(const uint16_t& a_rcv_sequence_number){
        if ((last_received_seq_number_ == -1) || (a_rcv_sequence_number == 0)){
            // start, just reset and record
            reset();
            //dialog_showed_ = false;
            return;
        }
        int diff_ = int(a_rcv_sequence_number) - int(last_received_seq_number_);
        if (diff_ > 1){
            //lost packet error
            // if (packet_drops_count_ > MAX_LOST_PACKETS_GLOBAL){
            //     doCBunError("To many UDP packets lost in total! : "+ std::to_string(packet_drops_count_));
            //     packet_drops_count_ = 0;
            // }
            global_packets_lost_seq += (diff_ - 1);
        }
        // if (diff_ < 0){
        //     doCBunError("Negative UDP swap found: "+ std::to_string(diff_));
        // }
    }

        // checks sequence numbers continuity (on STATS_INTERVAL_)
    void checkOfTheReceivedSequenceNumberLocal(const uint16_t& a_rcv_sequence_number){
        if ((last_received_seq_number_ == -1) || (a_rcv_sequence_number == 0)){
            // start, just record
            reset();
            //dialog_showed_ = false;
            return;
        }
        int diff_ = int(a_rcv_sequence_number) - int(last_received_seq_number_);
        if (diff_ > 1){ // ok difference is 1
            if ((diff_ - 1) > stats_parameters_.MAX_CONSECUTIVE_COMMANDS_LOST){
                doCBunError(network_alert_prefix + ": " + std::to_string((diff_ - 1)) + " packets lost in a row over " + doubleToStringWithPrecision(stats_parameters_.STATS_INTERVAL / 1000.) + " second/s.");
            }
            local_packets_lost_seq += (diff_ - 1);
            stats_buffer_seq.push((diff_ - 1));
        }
        else{
            stats_buffer_seq.push(0);
        }
        if (stats_buffer_seq.size() > stats_parameters_.STATS_INTERVAL / KORD_PERIOD_){ // STATS_INTERVAL_ passed
            if (local_packets_lost_seq > stats_parameters_.MAX_RECENT_COMMANDS_LOST){
                doCBunError(network_alert_prefix + ": " + std::to_string(local_packets_lost_seq) + " packets lost over " + doubleToStringWithPrecision(stats_parameters_.STATS_INTERVAL / 1000.) + " second/s.");
            }
            local_packets_lost_seq -= stats_buffer_seq.front();
            stats_buffer_seq.pop();
        }
    }

    void stopRobot(){
        CLOG_INFO("Stopping robot...");
        kr2rc_api::Controller::CmdUserConsentParams consent_params;

        std::string rc_command = "ctl/suspend"; 
        consent_params.data_ = rc_command.c_str();
        consent_params.data_len_ = rc_command.length();

        program_api_->rc_api_->system_controller_->cmd_UserConsent(consent_params);
        CLOG_INFO("Sent the stopping command!");
    }

    void clear( std::queue<int> &q ){
        std::queue<int> empty;
        std::swap( q, empty );
    }


    bool isMoveContinous(const kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType& a_track_type, 
        const double& a_track_value, const kr2rc_api::Trajectory::CmdTrajectoryParams::BlendType& a_blend_type, const double& a_blend_value){
        return (a_track_type == kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType::TT_TIME) &&
               (a_track_value == 0.008)                                                            &&
               (a_blend_type == kr2rc_api::Trajectory::CmdTrajectoryParams::BlendType::BT_TIME)    &&
               (a_blend_value == 0.004);
    }

    // Move ended with 0 velocity
    bool isMoveEndedRight(bool is_sum_coord_type_joint, const kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType& a_track_type, 
        const double& a_track_value){
        double epsilon = 1e-6;
        if (is_sum_coord_type_joint){
            epsilon = stats_parameters_.JOINT_MAX_DIFF_WHEN_MOVE_ENDED;
        }
        else{ // linear
            epsilon = stats_parameters_.LINEAR_MAX_DIFF_WHEN_MOVE_ENDED;
        }

        switch (a_track_type)
        {
        case kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType::TT_TIME:
            return (last_diff < epsilon);
            break;
        case kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType::TT_WS_TARGET_SPEED:
        case kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType::TT_JS_TARGET_SPEED:
            return (a_track_value > -1 * epsilon) && (a_track_value < epsilon);
        default:
            break;
        }
    }

    std::chrono::high_resolution_clock::time_point timespec_to_chrono (timespec a_t){
        std::chrono::high_resolution_clock::time_point ret{std::chrono::seconds(a_t.tv_sec) + std::chrono::nanoseconds(a_t.tv_nsec)};
        return ret;
    }

    int64_t tick_delay_micro(std::chrono::high_resolution_clock::time_point a_tick, std::chrono::high_resolution_clock::time_point a_curr_time){
        return std::chrono::duration_cast<std::chrono::microseconds>(a_curr_time - a_tick).count();
    }

    std::string doubleToStringWithPrecision(double number, int precision = 2) {
        std::stringstream stream;
        stream << std::fixed << std::setprecision(precision) << number;
        return stream.str();
    }


    // checks timestamps and calculates the lost number based on them
    void calcLostOnTimestamps(const int64_t& a_delay){
        int diff_ = a_delay / (KORD_PERIOD_ *  MICROSECS_IN_MILLI_);
        if (diff_ > 1){ // ok difference is 1
            if ((diff_ - 1) > stats_parameters_.MAX_CONSECUTIVE_COMMANDS_LOST){
                doCBunError(network_alert_prefix + ": " + std::to_string((diff_ - 1)) + " packets lost in a row over " + doubleToStringWithPrecision(stats_parameters_.STATS_INTERVAL / 1000.) + " second/s.");
            }
            local_packets_lost_timestmp += (diff_ - 1);
            global_packets_lost_timestmp += (diff_ - 1);
            stats_buffer_times.push((diff_ - 1));
        }
        else{
            stats_buffer_times.push(0);
        }
        if (stats_buffer_times.size() > stats_parameters_.STATS_INTERVAL / KORD_PERIOD_){ // STATS_INTERVAL_ passed
            if (local_packets_lost_timestmp > stats_parameters_.MAX_RECENT_COMMANDS_LOST){
                doCBunError(network_alert_prefix + ": " + std::to_string(local_packets_lost_timestmp) + " packets lost over " + doubleToStringWithPrecision(stats_parameters_.STATS_INTERVAL / 1000.) + " second/s.");
            }
            local_packets_lost_timestmp -= stats_buffer_times.front();
            stats_buffer_times.pop();
        }
    }


    // calculation of the Round trip times and commands delays
    bool calculateCaptureStats(bool is_sum_coord_type_joint, double sum_coord_diff, std::chrono::_V2::system_clock::time_point a_recent_capt_timestamp, timespec a_recent_capt_content_timestamp,
            kr2rc_api::Trajectory::CmdTrajectoryParams::TrackingType a_track_type, double a_track_value,
            kr2rc_api::Trajectory::CmdTrajectoryParams::BlendType a_blend_type, double a_blend_value){
        // for each in accum calc and check
        // updateSTickDelay lambda
        auto _updateSTickDelay = [this](timespec t_) -> int64_t
        {
            timespec now_ts                                         = t_;
            timespec current_tick                                   = kr2rc_api_internals::Timex::thisTick(now_ts, 0 );
            std::chrono::high_resolution_clock::time_point now_tick = timespec_to_chrono(current_tick);
            return tick_delay_micro(now_tick, timespec_to_chrono(now_ts));
        };

        // ltick
        int64_t     current_delay       = 0;
        int64_t     current_jitter      = 0;
        bool first_time = (last_timestamp_ == std::chrono::high_resolution_clock::time_point{});
        //statistics is only on continous moves
        if (isMoveContinous(a_track_type, a_track_value, a_blend_type, a_blend_value)){
            if (!first_time){
                current_delay = std::chrono::duration_cast<std::chrono::microseconds>(a_recent_capt_timestamp - last_timestamp_).count();
                //CLOG_INFO("current_delay: " << current_delay);
                if (current_delay > 100 * KORD_PERIOD_ * MICROSECS_IN_MILLI_){
                    // continous move ended, it's a start of a new continous move
                    accum_ltick.reset();
                    accum_stick.reset();
                    notified_about_velocity = false;
                }
                else{
                    if (current_delay == 0){ // end of a row of commands or connection lost
                        if (!notified_about_velocity && !isMoveEndedRight(is_sum_coord_type_joint, a_track_type, a_track_value)){
                            doCBunError(network_alert_prefix + ": Commands break down.");
                            notified_about_velocity = true;
                            return false;
                        }
                    }

                    // real delay between two consecutive continous moves
                    calcLostOnTimestamps(current_delay);

                    current_jitter = current_delay - KORD_PERIOD_ * MICROSECS_IN_MILLI_;
                    accum_ltick.measurement(current_jitter);
                    accum_ltick.compute();
                    if (accum_ltick.get_avg() > stats_parameters_.MAX_RECENT_AVG_CMD_JITTER){
                        // stop and notify
                        doCBunError(network_alert_prefix + ": " + "Average Local Command Jitter Exceeded: " + doubleToStringWithPrecision(accum_ltick.get_avg()) + " us, "
                         + "LIMIT: " + std::to_string(stats_parameters_.MAX_RECENT_AVG_CMD_JITTER) + " us.");
                    }

                    if (accum_ltick.get_max() > global_ltick_max){
                        global_ltick_max = accum_ltick.get_max();
                    }

                    // content tick
                    int64_t tick_delay = _updateSTickDelay(a_recent_capt_content_timestamp);
                    accum_stick.measurement(tick_delay);
                    accum_stick.compute();

                    if (accum_stick.get_avg() > stats_parameters_.MAX_RECENT_AVG_ROUNDTRIP_TIME){
                        doCBunError(network_alert_prefix + ": " + "Average Local Round Trip Time Exceeded: " + doubleToStringWithPrecision(accum_stick.get_avg()) + " us, "
                        + "LIMIT: " + std::to_string(stats_parameters_.MAX_RECENT_AVG_ROUNDTRIP_TIME) + " us.");
                    }

                    if (accum_stick.get_max() > global_stick_max){
                        global_stick_max = accum_stick.get_max();
                    }

                }
            }

            last_timestamp_ = a_recent_capt_timestamp;
        }
        else{
            accum_ltick.reset();
            accum_stick.reset();
            last_timestamp_ = std::chrono::high_resolution_clock::time_point{};
        }


        return true;
    }

    public:
        // for captured statistics, to send to API
        // local
        Accum<int64_t> accum_ltick; // command jitter  - micros
        Accum<int64_t> accum_stick; // round trip time - micros
        // System jitter
        Accum<int64_t> accum_wake_up_stats; // micros

        int64_t local_packets_lost_seq = 0;
        int64_t local_packets_lost_timestmp = 0;

        // global
        int64_t global_ltick_max = 0; // command jitter global MAX - micros
        int64_t global_packets_lost_seq = 0; // commands lost global counter based on sequence numbers
        int64_t global_packets_lost_timestmp = 0; // commands lost global counter based on time differences
        int64_t global_system_jitter_max = 0; // system jitter global MAX - micros

        int64_t global_stick_max = 0; // global round trip time max

        bool sys_problem = false;

};

#endif